import { JSDOM } from "jsdom";
const crypto = require("crypto");
import MIMEParser from "@vereign/mime-parser";
import { expect, test } from "@jest/globals";
import { DOM } from "@vereign/dom";

import { PlainNormalizer, HTMLNormalizer } from "../../src";
import { listDirectories, readFile } from "./fileReader";
import { diffStringsUnified } from "jest-diff";

const SENT_EML_NAME = "sent.eml";
const RECEIVED_EML_NAME = "received.eml";

/**
 * Substitutes cid:* with base64 strings of the corresponding attachments
 */
const POPULATE_ATTACHMENTS = true;

interface CasesFilteringOptions {
  ignore?: string[];
  checkExclusively?: string[];
}

/**
 * Hashes base64 of the attachments to improve performance and readability
 * @param base64
 */
const hashAttachment = (base64: string) => {
  return crypto
    .createHash("sha256")
    .update(Buffer.from(base64, "base64"))
    .digest()
    .toString("base64");
};

expect.extend({
  /**
   * Extends "toEqual" with diff functionality
   */
  toEqualWithDiff(target, source) {
    let pass = true;
    try {
      expect(target).toEqual(source);
    } catch (e) {
      pass = false;
    }

    return {
      pass,
      message: pass ? () => "Pass" : () => diffStringsUnified(source, target),
    };
  },
});

export const createHtmlTestsRunner = (testsPath: string, vendor: string) => (
  casesGroupName: string,
  filteringOptions?: CasesFilteringOptions
) => (): void => {
  const testCases = prepareTestCases(
    testsPath,
    casesGroupName,
    filteringOptions
  );

  test.concurrent.each(testCases)(
    "Case %s",
    async (dirName, { sentMime, receivedMime }) => {
      const { sentHtmlDocument, receivedHtmlDocument } = await getDOMDocuments(
        sentMime,
        receivedMime
      );

      const sentHtml = HTMLNormalizer.normalizeVendorHtml(
        sentHtmlDocument,
        vendor
      );
      const receivedHtml = HTMLNormalizer.normalizeVendorHtml(
        receivedHtmlDocument,
        vendor
      );

      // eslint-disable-next-line
      // @ts-ignore
      expect(receivedHtml).toEqualWithDiff(sentHtml);
    }
  );
};

export const createPlainTestsRunner = (testsPath: string) => (
  casesGroupName: string,
  filteringOptions?: CasesFilteringOptions
) => (): void => {
  const testCases = prepareTestCases(
    testsPath,
    casesGroupName,
    filteringOptions
  );

  test.concurrent.each(testCases)(
    "Case %s",
    async (dirName, { sentMime, receivedMime }) => {
      const sentNormalizedPlain = PlainNormalizer.normalizePlain(
        await sentMime.getPlain({
          populateAttachments: POPULATE_ATTACHMENTS,
          hashAttachment,
        })
      );
      const receivedNormalizedPlain = PlainNormalizer.normalizePlain(
        await receivedMime.getPlain({
          populateAttachments: POPULATE_ATTACHMENTS,
          hashAttachment,
        })
      );

      // eslint-disable-next-line
      // @ts-ignore
      expect(receivedNormalizedPlain).toEqualWithDiff(sentNormalizedPlain);
    }
  );
};

export const createPseudoplainTestsRunner = (
  testsPath: string,
  vendor: string
) => (
  casesGroupName: string,
  filteringOptions?: CasesFilteringOptions
) => (): void => {
  const testCases = prepareTestCases(
    testsPath,
    casesGroupName,
    filteringOptions
  );

  test.concurrent.each(testCases)(
    "Case %s",
    async (dirName, { sentMime, receivedMime }) => {
      const { sentHtmlDocument, receivedHtmlDocument } = await getDOMDocuments(
        sentMime,
        receivedMime
      );

      HTMLNormalizer.normalizeVendorHtml(sentHtmlDocument, vendor);
      HTMLNormalizer.normalizeVendorHtml(receivedHtmlDocument, vendor);

      const normalizedSentPseudoPlainText = HTMLNormalizer.extractPseudoPlainPart(
        sentHtmlDocument
      );

      const normalizedReceivedPseudoPlainText = HTMLNormalizer.extractPseudoPlainPart(
        receivedHtmlDocument
      );

      // eslint-disable-next-line
      // @ts-ignore
      expect(normalizedReceivedPseudoPlainText).toEqualWithDiff(
        normalizedSentPseudoPlainText
      );
    }
  );
};

const prepareTestCases = (
  path: string,
  casesGroup: string,
  filteringOptions: CasesFilteringOptions
): Array<[string, { sentMime: MIMEParser; receivedMime: MIMEParser }]> => {
  const testsCasesPath = path + "/" + casesGroup;
  const casesDirs = listDirectories(testsCasesPath);
  const casesDirsFiltered = filterTestCases(casesDirs, filteringOptions);

  return casesDirsFiltered.map((caseDir) => {
    const sentMimeString = readFile(
      `${testsCasesPath}/${caseDir}/${SENT_EML_NAME}`
    );
    const receivedMimeString = readFile(
      `${testsCasesPath}/${caseDir}/${RECEIVED_EML_NAME}`
    );

    return [
      caseDir,
      {
        sentMime: new MIMEParser(sentMimeString),
        receivedMime: new MIMEParser(receivedMimeString),
      },
    ];
  });
};

/**
 *
 * @param {Array<string>} casesNames
 * @param {Object} options
 * @param {Array<string>}  options.ignore - ignores specific test cases
 * @param {Array<string>}  options.checkExclusively - focuses on specific test cases (disregards toIgnore)
 */
const filterTestCases = (
  casesNames: string[],
  options: CasesFilteringOptions
): string[] => {
  const ignore = options?.ignore;
  const checkExclusively = options?.checkExclusively;

  if (checkExclusively?.length) {
    return casesNames.filter((caseName) => checkExclusively.includes(caseName));
  }

  if (ignore?.length) {
    return casesNames.filter((caseName) => !ignore.includes(caseName));
  }

  return casesNames;
};

export const getDOMDocuments = async (
  sentMime: MIMEParser,
  receivedMime: MIMEParser
): Promise<{
  sentHtmlDocument: HTMLDocument;
  receivedHtmlDocument: HTMLDocument;
}> => {
  /**
   * Usage of the @vereign/dom is a must, because of Google Add-on.
   */
  const sentDOM = new DOM(
    await sentMime.getHTML({
      populateAttachments: POPULATE_ATTACHMENTS,
      hashAttachment,
    })
  );
  const receivedDOM = new JSDOM(
    await receivedMime.getHTML({
      populateAttachments: POPULATE_ATTACHMENTS,
      hashAttachment,
    })
  );

  return {
    sentHtmlDocument: sentDOM.window.document,
    receivedHtmlDocument: receivedDOM.window.document,
  };
};
